
import imaplib
import email
import traceback
import enum
import os
import re
from email.header import decode_header, make_header

from typing import Dict, Optional, Tuple, Union
from bs4 import BeautifulSoup
import pandas as pd


imaplib.Commands['ID'] = ('AUTH',)

UID_FILE_NAME = 'mail_uid_recorder'


class IMAP_SPType(enum.IntEnum):
    IMAP_QQ = 1  # 企业微信邮箱


class MailAttachmentRetriever:

    _IMAP_SERVER_INFO = {
        IMAP_SPType.IMAP_QQ: ('imap.exmail.qq.com', 993)
    }

    def __init__(self, dump_dir: str, file_suffixes: Tuple[str]):
        self._dump_dir = dump_dir
        assert os.path.isdir(self._dump_dir), f'arg dump_dir should be a directory (now){self._dump_dir}'

        self._file_suffixes = []
        for one in file_suffixes:
            if not one.startswith('.'):
                self._file_suffixes.append('.' + one)
            else:
                self._file_suffixes.append(one)
        print(f'(file suffixes){self._file_suffixes}')

    def get_excels(self, sp_type_or_uri: Union[IMAP_SPType, str], user: str, pw: str, last_uid: bytes = None) -> Optional[Dict[str, Tuple[bytes, os.PathLike]]]:
        if isinstance(sp_type_or_uri, IMAP_SPType):
            try:
                host, port = self._IMAP_SERVER_INFO[sp_type_or_uri]
            except KeyError:
                print(f'invalid SP type {sp_type_or_uri}, do not support it')
                return
        else:
            try:
                host, port = sp_type_or_uri.split(':')
            except Exception:
                print(f'invalid sp uri {sp_type_or_uri}, the format should be url:port')
                return
        try:
            df_list: Dict[str, pd.DataFrame] = {}
            with imaplib.IMAP4_SSL(host=host, port=port) as M:
                M.login(user, pw)
                args = ("name", "PuyuanTech", "contact", "PuyuanTech", "version", "1.0.0", "vendor", "myclient")
                id_arg = '" "'.join(args)
                typ, dat = M._simple_command('ID', f'("{id_arg}")')
                # print(M._untagged_response(typ, dat, 'ID'))
                ret = M.select(mailbox='Inbox', readonly=True)
                print(f'(select ret){ret}')
                if last_uid is None:
                    criterion = 'ALL'
                else:
                    last_uid = int(last_uid)
                    # criterion = f'(TO "fof@puyuan.tech" UID {last_uid}:*)'
                    criterion = f'(UID {last_uid}:*)'
                    print(criterion)
                dummy_count = 0
                typ, data = M.uid('search', None, criterion)
                for uid in data[0].split():
                    # 上边的criterion不太好用 这里还是需要再过滤一下
                    if last_uid is not None and int(uid) <= last_uid:
                        continue
                    # 邮件应该是按顺序过来的
                    # 遍历每一封邮件
                    is_done = False
                    typ, data = M.uid('fetch', uid, '(RFC822)')
                    raw_email = data[0][1]
                    email_message = email.message_from_bytes(raw_email)
                    for part in email_message.walk():
                        # excel类型的附件这里的content maintype其实都是application
                        if part.get_content_maintype() != 'multipart':
                            # 这个条件是判断附件的关键
                            if part.get('Content-Disposition') is not None:
                                filename = part.get_filename()
                                if filename is not None:
                                    the_header = email.header.decode_header(filename)
                                    print(f'header length: {len(the_header)}')
                                    for h in the_header:
                                        print(f'>> {h}')
                                    real_name = the_header[0][0]
                                    print(f'name before decode: {real_name} (type){type(real_name)}')
                                    print(f'codec: {the_header[0][1]}')
                                    if isinstance(real_name, bytes):
                                        real_name = real_name.decode(the_header[0][1])
                                        print(f'name after decode: {real_name}')
                                    dummy_count += 1
                                    for one in self._file_suffixes:
                                        if real_name.endswith(one):
                                            # pandas可以直接读，但pd.read_excel()里传的参数可能是不一样的，所以这里还是先将内容写到文件里
                                            real_name = f'c{dummy_count}_{real_name}'
                                            file_path = os.path.join(self._dump_dir, real_name)
                                            with open(file_path, 'wb') as f:
                                                f.write(part.get_payload(decode=True))
                                            df_list[real_name] = (uid, os.path.abspath(file_path))
                                            print(f'file {real_name} on {df_list[real_name]} done (uid){uid} (dummy_count){dummy_count}')
                                            is_done = True
                                            break
                    if not is_done:
                        for part in email_message.walk():
                            if part.get('Content-Disposition') is None:
                                # 尝试解析正文
                                body = part.get_payload(decode=True)
                                if body is None:
                                    continue
                                soup = BeautifulSoup(body, features='lxml')
                                if soup.table is None:
                                    continue

                                for one in soup.findAll('meta', {'content': re.compile(r'.*charset.*')}):
                                    encoding = re.findall(r'.*charset=(.*)', one.get('content'))
                                    if encoding:
                                        encoding = encoding[0]
                                        break
                                else:
                                    encoding = 'gbk'
                                try:
                                    df = pd.read_html(str(soup.table), encoding=encoding)
                                except Exception as e:
                                    print(f'got exception when read html (exception){e}')
                                else:
                                    if df:
                                        df = df[0]
                                        subject = str(make_header(decode_header(email_message['Subject'])))
                                        fund_id = re.findall(r'S[A-Z0-9]{5}', subject, flags=re.ASCII)
                                        if fund_id and '产品代码' not in df.columns.array:
                                            df['产品代码'] = fund_id[0]
                                        real_name = f"{subject}_{encoding}.xlsx"
                                        file_path = os.path.join(self._dump_dir, real_name)
                                        try:
                                            if df.shape[0] < 2:
                                                df.to_excel(file_path, index=False)
                                            else:
                                                df.to_excel(file_path, header=None, index=False)
                                            df_list[real_name] = (uid, os.path.abspath(file_path))
                                            print(f'content {real_name} on {df_list[real_name]} done (uid){uid}')
                                        except Exception:
                                            pass
                M.close()
            return df_list
        except Exception as e:
            print(e)
            traceback.print_exc()


if __name__ == '__main__':
    try:
        email_data_dir = os.environ['SURFING_EMAIL_DATA_DIR']
        user_name = os.environ['SURFING_EMAIL_USER_NAME']
        password = os.environ['SURFING_EMAIL_PASSWORD']
    except KeyError as e:
        import sys
        sys.exit(f'can not found enough params in env (e){e}')

    mar = MailAttachmentRetriever(email_data_dir, ['xls', 'xlsx', 'pdf'])
    print(mar.get_excels(IMAP_SPType.IMAP_QQ, user_name, password, b'926'))




import os
import dataclasses
import time
from typing import List, Optional
from collections import defaultdict

import pandas as pd

from ..._version import __version__
from ...constant import SectorType, IndexPriceSource
from ...util.config import SurfingConfigurator, AWSSettings
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from ..view.raw_models import CSIndexComponent
from ..view.basic_models import FundInfo, TradingDayList, IndexInfo, IndexPrice, FundNav, Fund_size_and_hold_rate, FundBenchmark, FundIPOStats
from ..view.basic_models import FundConvStats, FundStatusLatest, FundOpenInfo
from ..view.derived_models import FundIndicator, FundIndicatorAnnual, IndexValuationLongTerm, FundIndicatorWeekly
from ..view.derived_models import FundIndicatorMonthly, BarraCNE5FactorReturn, FundManagerScore, FundManagerInfo, FundManagerFundRank
from .data_tables import FundDataTables
from .score import FundScoreManager


_DEFAULT_BUCKET_NAME = 'tl-fund-dm'

# 存储版本号文件的文件名
_VERSION_FILE_NAME = '_version.csv'


class S3ETLTool:
    def __init__(self):
        try:
            self._conf = SurfingConfigurator().get_aws_settings()
        except AssertionError as e:
            print(f'{e}, use default value')
            self._conf = AWSSettings.get_default_config()
        print('s3 config: ')
        print(f'    {self._conf}')
        if self._conf.etl_tool_bucket_name is None or not self._conf.etl_tool_bucket_name:
            bucket_name = _DEFAULT_BUCKET_NAME
        else:
            bucket_name = self._conf.etl_tool_bucket_name
        # 原始(指从数据库里读出来未加以任何调整的)table的uri on s3
        self._raw_s3_bucket_uri: str = f's3://{bucket_name}/raw_tables'
        # dm调整后的data的uri on s3
        self._s3_bucket_uri: str = f's3://{bucket_name}/dm'
        print('bucket uri: ')
        print(f'    {self._raw_s3_bucket_uri}')
        print(f'    {self._s3_bucket_uri}')
        # 在to_parquet/read_parquet中通过storage_options传递如下参数的方法不好用，这里直接设置环境变量
        os.environ['AWS_ACCESS_KEY_ID'] = self._conf.aws_access_key_id
        os.environ['AWS_SECRET_ACCESS_KEY'] = self._conf.aws_secret_access_key
        os.environ['AWS_DEFAULT_REGION'] = self._conf.region_name


class SynchronizerS3ForDM(S3ETLTool):
    '''将数据同步到S3的工具'''

    # DM中需要使用的raw table
    _RAW_TABLE_LIST = [
        CSIndexComponent,
    ]
    # DM中需要使用的basic table
    _BASIC_TABLE_LIST = [
        FundInfo,
        TradingDayList,
        IndexInfo,
        IndexPrice,
        FundNav,
        Fund_size_and_hold_rate,
        FundBenchmark,
        FundIPOStats,
        FundConvStats,
        FundStatusLatest,
        FundOpenInfo,
    ]
    # DM中需要使用的derived table
    _DERIVED_TABLE_LIST = [
        FundIndicator,
        FundIndicatorAnnual,
        IndexValuationLongTerm,
        FundIndicatorWeekly,
        FundIndicatorMonthly,
        BarraCNE5FactorReturn,
        FundManagerScore,
        FundManagerInfo,
        FundManagerFundRank,
    ]

    # 如果不传dm, 可以使用dump_all_tables
    # 否则也可以使用dump_all_dfs
    def __init__(self, dm=None):
        super().__init__()

        if dm is not None:
            # 存下来dts中所有的字段和值
            self._data_tables_dict = dataclasses.asdict(dm.dts)
            # 存下来dm的score manager
            self._score_manager = dm._score_manager

    def _dump_dts(self) -> List[str]:
        '''dump all dfs in dts'''

        failed_tables: List[str] = []
        for name, data in self._data_tables_dict.items():
            if not isinstance(data, pd.DataFrame) or (data is None):
                # 这里不处理df以外的数据
                continue

            _t0 = time.time()
            try:
                # if name == 'cs_index_component':
                #     data['id_cat'] = data.id_cat.transform(lambda x: x.value)
                # elif name == 'index_info':
                #     data['price_source'] = data.price_source.transform(lambda x: x.value)
                data.to_parquet(f'{self._s3_bucket_uri}/{name}.parquet', compression='gzip')
            except Exception as e:
                print(f'WARNING!! dump df {name} to s3 failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, table {name} to s3 done')
        return failed_tables

    def _dump_misc(self) -> List[str]:
        '''dump all datas except dfs in dts and all datas in score manager'''

        # dts部分
        failed_tables: List[str] = []
        for name, data in self._data_tables_dict.items():
            if isinstance(data, pd.DataFrame) or (data is None):
                # 这里只处理df以外的数据
                continue
            if name in ('mng_best_fund', 'mng_index_list'):
                # 这两个data是在fund_manager_processor_advanced里dump的
                continue

            # 由于df写parquet很方便，因此这里我们做一些类型转换，把类型均转为df
            if isinstance(data, set) or isinstance(data, list):
                df = pd.DataFrame.from_dict({name: list(data)})
            elif isinstance(data, dict):
                df = pd.Series(data).to_frame(name)
            else:
                print(f'ERROR!! do not recognize the type of data (type){type(data)} (name){name}')
                failed_tables.append(name)
                continue

            _t0 = time.time()
            try:
                df.to_parquet(f'{self._s3_bucket_uri}/_ndf_{name}.parquet', compression='gzip')
            except Exception as e:
                print(f'WARNING!! dump data {name} to s3 failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, data {name} to s3 done')

        # score manager部分
        for name in ('score_cache', 'score_raw_cache', 'score_manager_cache'):
            _t0 = time.time()
            try:
                df = pd.Series(self._score_manager.__dict__[name]).to_frame(name)
                if name == 'score_manager_cache':
                    df['score_manager_cache'] = df.score_manager_cache.transform(lambda x: {k: v.to_dict() for k, v in x.items()})
                df.to_parquet(f'{self._s3_bucket_uri}/_ndf_{name}.parquet', compression='gzip')
            except Exception as e:
                print(f'WARNING!! dump data {name} to s3 failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, , data {name} to s3 done')
        return failed_tables

    def _dump_whole_table_data(self, session, table_list, processor=None) -> List[str]:
        '''dump whole data of tables from db'''

        def _fetch_whole_table(session, view):
            query = session.query(view)
            return pd.read_sql(query.statement, query.session.bind)

        failed_tables: List[str] = []
        for table in table_list:
            name = table.__table__.name
            try:
                print(f'to dump table {name}')
                _t0 = time.time()
                data = _fetch_whole_table(session, table)
                if processor is not None:
                    data = processor(table, data)
                _t1 = time.time()
                print(f'cost {_t1 - _t0}s, and send table {name} to s3')
                data.to_parquet(f'{self._raw_s3_bucket_uri}/{name}.parquet', compression='gzip', index=False)
                _t2 = time.time()
                print(f'cost {_t2 - _t1}s, table {name} to s3 done')
            except Exception as e:
                print(f'WARNING!! dump table {name} failed (err_msg){e}')
                failed_tables.append(name)
        return failed_tables

    def _dump_raw_table(self) -> List[str]:
        def _raw_table_processor(table, data):
            # if table == CSIndexComponent:
            #     data['id_cat'] = data.id_cat.transform(lambda x: x.value)
            return data

        print('to dump raw tables')
        with RawDatabaseConnector().managed_session() as session:
            failed_tables: List[str] = self._dump_whole_table_data(session, self._RAW_TABLE_LIST, _raw_table_processor)
        print('dump raw tables done')
        return failed_tables

    def _dump_basic_table(self) -> List[str]:
        def _basic_table_processor(table, data):
            # if table == IndexInfo:
            #     data['price_source'] = data.price_source.transform(lambda x: x.value)
            return data

        print('to dump basic tables')
        with BasicDatabaseConnector().managed_session() as session:
            failed_tables: List[str] = self._dump_whole_table_data(session, self._BASIC_TABLE_LIST, _basic_table_processor)
        print('dump basic tables done')
        return failed_tables

    def _dump_derived_table(self) -> List[str]:
        print('to dump derived tables')
        with DerivedDatabaseConnector().managed_session() as session:
            failed_tables: List[str] = self._dump_whole_table_data(session, self._DERIVED_TABLE_LIST)
        print('dump derived tables done')
        return failed_tables

    def dump_one_df(self, name: str) -> bool:
        _t0 = time.time()
        try:
            data = self._data_tables_dict[name]
        except KeyError as e:
            print(e)
            return False

        if not isinstance(data, pd.DataFrame):
            # 这里不处理df以外的数据
            print(f'{name} is not a df, can not dump it')
            return False

        try:
            # if name == 'cs_index_component':
            #     data['id_cat'] = data.id_cat.transform(lambda x: x.value)
            # elif name == 'index_info':
            #     data['price_source'] = data.price_source.transform(lambda x: x.value)
            data.to_parquet(f'{self._s3_bucket_uri}/{name}.parquet', compression='gzip')
        except Exception as e:
            print(f'WARNING!! dump df {name} to s3 failed (err_msg){e}')
        _t1 = time.time()
        print(f'cost {_t1 - _t0}s, table {name} to s3 done')
        return True

    def dump_all_dfs(self) -> List[str]:
        try:
            pd.DataFrame({'version': __version__}, index=['now']).to_csv(f'{self._s3_bucket_uri}/{_VERSION_FILE_NAME}')
            print(f'version {__version__} to s3 done')
        except Exception as e:
            print(f'WARNING!! version {__version__} to s3 failed (err_msg){e}')
            return ['version']

        failed_tables: List[str] = self._dump_dts()
        failed_tables += self._dump_misc()
        print(f'dump all dfs done (failed tables){failed_tables}')
        return failed_tables

    def dump_all_tables(self) -> List[str]:
        try:
            pd.DataFrame({'version': __version__}, index=['now']).to_csv(f'{self._raw_s3_bucket_uri}/{_VERSION_FILE_NAME}')
            print(f'version {__version__} to s3 done')
        except Exception as e:
            print(f'WARNING!! version {__version__} to s3 failed (err_msg){e}')
            return ['version']

        failed_tables: List[str] = self._dump_raw_table()
        failed_tables += self._dump_basic_table()
        failed_tables += self._dump_derived_table()
        print(f'dump all tables done (failed tables){failed_tables}')
        return failed_tables


class S3RetrieverForDM(S3ETLTool):
    '''从S3获取数据的工具'''

    def __init__(self, activation: Optional[List[str]] = None):
        super().__init__()
        # 默认先不激活
        self._activation: S3RetrieverForDM = activation

    @property
    def is_activated(self) -> bool:
        # 只有self._activation是空列表的情况才返回True; 非空表示一部分table需要从mysql读取
        return self._activation is not None and not self._activation

    def fetch_table(self, query, post_processor=None):
        table_name = query.selectable.froms[0].name
        # 如果没有开启的话，直接从mysql读取
        if self._activation is None or (table_name in self._activation):
            print(f'loading data of {table_name} from sql')
            return pd.read_sql(query.statement, query.session.bind).drop(columns='_update_time', errors='ignore')

        try:
            print(f'loading data of {table_name} from s3')
            df = pd.read_parquet(f'{self._raw_s3_bucket_uri}/{table_name}.parquet').drop(columns='_update_time', errors='ignore')
            columns_need = []
            col_desc = query.column_descriptions
            for one in col_desc:
                # 这里表明是查询全表，而不是只查询特定几列
                if one['type'] == one['entity']:
                    break
                # 把查询的特定几列的列名字都记下来
                columns_need.append(one['name'])
            else:
                # 只取查询的特定几列
                df = df[columns_need]

            if post_processor is not None:
                df = post_processor(df)

            # 这里为了与从DB里读出来的数据内容一样，需要做一个转换
            if table_name == 'cs_index_component':
                df['id_cat'] = df.id_cat.transform(lambda x: SectorType(x))
            elif table_name == 'index_info':
                df['price_source'] = df.price_source.transform(lambda x: IndexPriceSource(x))
            return df
        except Exception as e:
            print(f'WARNING!! fetch table {table_name} from s3 failed (err_msg){type(e).__name__}: {e}, try mysql')
            # TODO: 如果这里DB也失败，应该也需要适当的处理
            return pd.read_sql(query.statement, query.session.bind).drop(columns='_update_time', errors='ignore')

    def retrieve_dts(self, data_tables: FundDataTables, candidates=None) -> List[str]:
        failed_tables: List[str] = []
        for field in dataclasses.fields(data_tables):
            # 这里不处理df以外的数据
            if field.type != pd.DataFrame:
                continue
            # 以下df理论上没必要保存，可以自己生成出来且较快
            if field.name in ['index_fund_indicator_pack','stock_ipo_data','conv_bond_ipo_data','stock_ipo_data_detail','conv_bond_ipo_data_detail','stock_ipo_info','conv_bond_ipo_info']:
                continue
            name = field.name

            if candidates and name not in candidates:
                continue

            _t0 = time.time()
            try:
                if name in ['mng_indicator_score', 'fund_indicator_score']:
                    _bucket_name = 'tl-fund-dm'
                    _s3_bucket_uri: str = f's3://{_bucket_name}/dm'
                    data = pd.read_parquet(f'{_s3_bucket_uri}/{name}.parquet')
                else:
                    data = pd.read_parquet(f'{self._s3_bucket_uri}/{name}.parquet')
                    # if name == 'cs_index_component':
                    #     data['id_cat'] = data.id_cat.transform()
                    # elif name == 'index_info':
                    #     data['price_source'] = data.price_source.transform()
                data_tables.__dict__[name] = data
            except (FileNotFoundError, PermissionError, Exception) as e:
                print(f'WARNING!! retrieve df {name} failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, retrieve df {name} from s3 done')
        return failed_tables

    def _retrieve_all_dts(self, data_tables: FundDataTables) -> List[str]:
        _t0 = time.time()
        failed_tables: List[str] = self.retrieve_dts(data_tables)
        _t1 = time.time()
        print(f'all dts total cost {_t1 - _t0}s (failed_tables){failed_tables}')
        return failed_tables

    def retrieve_misc(self, dm, candidates=None) -> List[str]:
        failed_tables: List[str] = []
        for field in dataclasses.fields(dm.dts):
            # 这里不处理df以外的数据
            if field.type == pd.DataFrame:
                continue
            name = field.name

            if candidates and name not in candidates:
                continue

            _t0 = time.time()
            try:
                # 把我们保存的df转换回原来的类型
                if name in ('fund_list', 'all_fund_list', 'fund_ipo_list', 'fund_conv_list', 'index_list', 'index_date_list'):
                    df = pd.read_parquet(f'{self._s3_bucket_uri}/_ndf_{name}.parquet')
                    data = set(df.to_dict(orient='list')[name])
                    if name == 'index_date_list':
                        data = sorted(data)
                elif name in ('mng_best_fund','mng_index_list'):
                    df = pd.read_parquet(f'{self._s3_bucket_uri}/{name}.parquet')
                    data = {}
                    for col in df:
                        _df = df[[col]]
                        data[col] = pd.DataFrame(_df[col].values.tolist(), index=_df.index)
                elif name in ('fund_index_map', 'fund_end_date_dict'):
                    df = pd.read_parquet(f'{self._s3_bucket_uri}/_ndf_{name}.parquet')
                    data = df.to_dict()[name]
                elif name in ('stock_ipo_data','conv_bond_ipo_data','stock_ipo_data_detail','conv_bond_ipo_data_detail','stock_ipo_info','conv_bond_ipo_info'):
                    data = {}
                else:
                    raise FileNotFoundError(f'do not support retrieve this data {name} from s3')
                dm.dts.__dict__[name] = data
            except (FileNotFoundError, PermissionError, Exception) as e:
                print(f'WARNING!! retrieve data {name} failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, retrieve data {name} from s3 done')
        return failed_tables

    def _retrieve_all_misc(self, dm) -> List[str]:
        _t0 = time.time()
        failed_tables: List[str] = self.retrieve_misc(dm)
        _t1 = time.time()
        print(f'all misc total cost {_t1 - _t0}s (failed_tables){failed_tables}')
        return failed_tables

    def retrieve_fund_score(self, score_manager: FundScoreManager):
        def remove_values_in_dict(item):
            if not isinstance(item, dict):
                return

            to_delete = []
            for k, v in item.items():
                if v is None:
                    to_delete.append(k)
                else:
                    remove_values_in_dict(v)
            for one in to_delete:
                del item[one]

        failed_tables: List[str] = []
        for name in ('score_cache', 'score_raw_cache', 'score_manager_cache'):
            _t0 = time.time()
            try:
                data = pd.read_parquet(f'{self._s3_bucket_uri}/_ndf_{name}.parquet').to_dict(into=defaultdict(dict))[name]
                # 转成df的过程中可能产生了很多None，这里把这些都干掉
                remove_values_in_dict(data)
                if name == 'score_manager_cache':
                    for v in data.values():
                        for lk, lv in v.items():
                            v[lk] = pd.Series(lv).rename_axis('manager_id')
                score_manager.__dict__[name] = data
            except (FileNotFoundError, PermissionError) as e:
                print(f'WARNING!! retrieve data {name} failed (err_msg){e}')
                failed_tables.append(name)
            _t1 = time.time()
            print(f'cost {_t1 - _t0}s, retrieve data {name} from s3 done')
        return failed_tables

    def retrieve_all_dfs(self, dm):
        try:
            version = pd.read_csv(f'{self._s3_bucket_uri}/{_VERSION_FILE_NAME}', index_col=0).at['now', 'version']
        except Exception as e:
            print(f'WARNING!! retrieve version info from s3 failed (err_msg){e}')
            return ['version']

        print(f'to retrieve all datas of version {version}')

        failed_datas: List[str] = self._retrieve_all_dts(dm.dts)
        failed_datas += self._retrieve_all_misc(dm)
        # failed_datas += self.retrieve_fund_score(dm._score_manager)
        return failed_datas



from numpy.lib.function_base import append
import pandas as pd
import numpy as np
import os
import datetime
import traceback
import json
import requests
import re
from bs4 import BeautifulSoup
import xml
from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util import Retry
from ...api.raw import RawDataApi
from ...view.raw_models import YahooIndexPrice, EmBTCFromWeb, EmFuturePriceDetail, EmFutureInfoDetail
from .raw_data_helper import RawDataHelper


class WebRawDataDownloader(object):

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self._session = Session()
        self._session.mount('https://', HTTPAdapter(
            max_retries=Retry(total=10, backoff_factor=0.02),
        ))
        self._session.mount('http://', HTTPAdapter(
            max_retries=Retry(total=10, backoff_factor=0.02),
        ))

    def read_csv(self, csv_path):
        return pd.read_csv(csv_path, index_col=1)

    def fund_fee(self):
        pass

    def wind_fund_info(self, csv_path):
        # Update manually
        # Default file in company wechat is 'fundlist_wind_20202015.xlsx'
        df = pd.read_excel(csv_path)
        df.columns = ['wind_id', 'desc_name', 'full_name', 'start_date', 'end_date', 'benchmark', 'wind_class_1',
            'wind_class_2', 'currency', 'base_fund_id', 'is_structured', 'is_open', 'manager_id', 'company_id']
        df['start_date'] = df['start_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d'))
        df['end_date'] = df['end_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d')
            if isinstance(x, str) else datetime.datetime.strptime('2040-12-31', '%Y-%m-%d'))
        df['is_structured'] = df['is_structured'].map(lambda x: False if x == '否' else True)
        df['is_open'] = df['is_open'].map(lambda x: False if x == '否' else True)
        self._data_helper._upload_raw(df, 'wind_fund_info')

    def cm_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 汇率数据.xls
            # df = pd.read_excel(csv_path)
            # df = df[['日期','美元中间价','欧元中间价','日元中间价','美元CFETS','欧元CFETS','日元CFETS']]
            # df.columns = ['datetime','usd_central_parity_rate','eur_central_parity_rate','jpy_central_parity_rate',
            #               'usd_cfets','eur_cfets','jpy_cfets']

            # auto update
            # data from http://www.chinamoney.com.cn/chinese/bkccpr
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
            url = ('http://www.chinamoney.com.cn/dqs/rest/cm-u-bk-ccpr/CcprHisExcelNew'
                f'?startDate={start_date}&endDate={end_date}&currency=USD/CNY,EUR/CNY,100JPY/CNY')
            headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}
            filename = ''
            try:
                r = self._session.get(url, headers=headers, timeout=(15, 60))
                filename = os.path.join('/tmp', 'cm.xlsx')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except Exception as e:
                print(f'[cm_index_price] got error {e}')
                return False
            df = pd.read_excel(filename)
            # Drop last 2 lines in excel, which are
            #   数据来源：	中国货币网
            #   www.chinamoney.com.cn
            df.drop(df.index[-2:], inplace=True)
            df = df[['日期', 'USD/CNY', 'EUR/CNY', '100JPY/CNY']]
            df.columns = ['datetime', 'usd_central_parity_rate', 'eur_central_parity_rate', 'jpy_central_parity_rate']
            df['jpy_central_parity_rate'] = df['jpy_central_parity_rate'] / 100
            df = df.sort_values('datetime')
            self._data_helper._upload_raw(df, 'cm_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def cxindex_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 中证信用债指数历史数据.xls
            # df = pd.read_excel(csv_path)
            # df.columns = ['index_id', 'symbol', 'datetime', 'open', 'high', 'low', 'close', 'total_turnover', 'volume']
            # df['ret'] = df['close'] / df['close'].shift(1) - 1
            # df['index_id'] = 'credit_debt'
            # df = df.drop(['symbol'], axis=1)[:-2]

            # auto update
            # data from http://www.csindex.com.cn/zh-CN/indices/index-detail/H11073
            # File download url
            url = 'http://www.csindex.com.cn/uploads/file/autofile/perf/H11073perf.xls'
            headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}
            filename = ''
            try:
                r = self._session.get(url, headers=headers, timeout=(15, 60))
                filename = os.path.join('/tmp', 'H11073perf.xls')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except Exception as e:
                print(f'[cxindex_index_price] got error {e}')
                return False
            df = pd.read_excel(filename)
            df = df[['日期Date', '收盘Close', '涨跌幅(%)Change(%)', '成交量（万元）Volume(10 thousand CNY)',
                '成交金额（元）Turnover']]
            df.columns = ['datetime', 'close', 'ret', 'volume', 'total_turnover']
            df['volume'] = df['volume'].values * 10000
            df['ret'] = df['ret'] / 100
            start_date = str(pd.to_datetime(start_date).date() - datetime.timedelta(days=10)).replace('-','')
            df = df[(df['datetime']>=start_date) & (df['datetime']<=end_date)].sort_values('datetime')
            df['open'] = float('Nan')
            df['high'] = float('Nan')
            df['low'] = float('Nan')
            df['index_id'] = 'credit_debt'
            data = RawDataApi().get_cxindex_index_price_df(start_date, end_date)
            lst_time = data['datetime'].tolist()
            lst_str = [str(x) for x in lst_time]
            time_all = df['datetime'].tolist()
            time_all = [str(x.date()) for x in time_all]
            time_add = list(set(time_all).difference(set(lst_str)))
            data_ = pd.DataFrame()
            for i in time_add:
                data_ = pd.concat([data_,df[ df['datetime'] == i]])
            self._data_helper._upload_raw(data_, 'cxindex_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def yahoo_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 大类资产指数据.xlsx
            # df = pd.read_excel(csv_path)
            # df['datetime'] = df['date']
            # for c in ['sp500', 'dax30', 'n225']:
            #     df_i = df[['datetime']].copy()
            #     df_i['close'] = df[c].copy()
            #     df_i['ret'] = df_i['close'] / df_i['close'].shift(1) - 1
            #     df_i['open'] = float('Nan')
            #     df_i['high'] = float('Nan')
            #     df_i['low'] = float('Nan')
            #     df_i['volume'] = float('Nan')
            #     df_i['total_turnover'] = float('Nan')
            #     df_i['index_id'] = c
            #     self._data_helper._upload_raw(df_i, 'yahoo_index_price')

            # auto update
            # data from
            # sp500 https://finance.yahoo.com/quote/%5EGSPC/history?p=%5EGSPC
            # n225 https://finance.yahoo.com/quote/%5EN225/history?p=%5EN225
            # dax30 https://finance.yahoo.com/quote/%5EGDAXI/history/
            # File download url:
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGSPC?period1=1553667756&period2=1585290156&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EN225?period1=1553673330&period2=1585295730&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGDAXI?period1=1553673346&period2=1585295746&interval=1d&events=history

            url_prefix = 'https://query1.finance.yahoo.com/v7/finance/download/'

            trading_day = RawDataApi().get_em_tradedates().set_index('TRADEDATES')
            start_date_loc = trading_day.index.get_loc(pd.to_datetime(start_date).date(), method='ffill')
            end_date_loc = trading_day.index.get_loc(pd.to_datetime(end_date).date(), method='ffill')

            assert isinstance(start_date_loc, int), 'start date is not a valid trading day'
            assert isinstance(end_date_loc, int), 'end date is not a valid trading day'

            # TODO: (deprecated)由于我们目前更新的时间（次日0点），美股还没收盘，所以这里只能更新前一天的数据；之后再看看可能需要分批做自动更新
            # TODO: 更新时间暂时统一挪到美股收盘后了，这里直接更新当天数据
            extra_days = 0
            assert start_date_loc >= extra_days, 'there is no enough extra days before start date'
            start_date_with_extra_days = trading_day.index.array[start_date_loc - extra_days]
            print(f'start date with extra days: {start_date_with_extra_days}')

            assert end_date_loc >= extra_days, 'there is no enough extra days before end date'
            end_date_with_extra_days = trading_day.index.array[end_date_loc - extra_days]
            print(f'end date with extra days: {end_date_with_extra_days}')

            start_date = min(pd.to_datetime(start_date, infer_datetime_format=True) - datetime.timedelta(days=7), start_date_with_extra_days)
            start_timestamp = str(int((start_date).timestamp()))
            end_timestamp = str(int((datetime.datetime.strptime(end_date, '%Y%m%d') + datetime.timedelta(days=1)).timestamp()))
            url_postfix = f'?period1={start_timestamp}&period2={end_timestamp}&interval=1d&events=history'
            headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}
            url_type_list = ['%5EGSPC', '%5EN225', '%5EGDAXI', '%5EFTSE', '%5EDJI', '%5Eixic']
            csv_list = ['^GSPC.csv', '^N225.csv', '^GDAXI.csv', '^FTSE.csv', '^DJI.csv', '^IXIC.csv']
            index_list = ['sp500', 'n225', 'dax30', 'ftse100', 'dji', 'ixic']
            df_list = []
            for url_type, csv_i, index_i in zip(url_type_list, csv_list, index_list):
                url = url_prefix + url_type + url_postfix
                filename = ''
                try:
                    r = self._session.get(url, headers=headers, timeout=(15, 60))
                    filename = os.path.join('/tmp', csv_i)
                    with open(filename, 'wb') as f:
                        f.write(r.content)
                except Exception as e:
                    print(f'[yahoo_index_price] got error {e}')
                    continue
                df = pd.read_csv(filename)
                df = df[['Date', 'Open', 'High', 'Low', 'Close', 'Adj Close', 'Volume']]
                df.columns = ['datetime', 'open', 'high', 'low', 'close', 'adjclose', 'volume']
                df['datetime'] = df['datetime'].map(lambda x: pd.to_datetime(x, infer_datetime_format=True).date())
                df['index_id'] = index_i
                df['total_turnover'] = float('Nan')
                df['ret'] = df['adjclose'] / df['adjclose'].shift(1) - 1
                df = df.drop(['adjclose'], axis=1)
                df = df[df['datetime'].between(start_date_with_extra_days, end_date_with_extra_days)].sort_values('datetime')
                if not df.empty:
                    df_list.append(df)
            if df_list:
                df_all = pd.concat(df_list)
                self._data_helper._upload_raw(df_all, YahooIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def sina_btc_download(self):
        try:
            url = 'https://stock2.finance.sina.com.cn/futures/api/jsonp.php/var%20pp=/GlobalFuturesService.getGlobalFuturesDailyKLine?symbol=BTC'
            response = requests.get(url)
            dic = response._content.decode('utf8').replace('/*<script>location.href=\'//sina.com\';</script>*/\nvar pp=(','')[:-2]
            data = json.loads(dic)
            data = pd.DataFrame(data)
            data = data[['date','close']].copy()
            data.loc[:,'codes'] = 'cme_btc_cfd'
            td = pd.to_datetime(data.date)
            td = [i.date() for i in td]
            data.date = td
            raw = RawDataApi()
            df_existed = raw.get_btc(['cme_btc_cfd'])
            data = data[~data.date.isin(df_existed.date.tolist())]
            self._data_helper._upload_raw(data, EmBTCFromWeb.__table__.name)
            return True
            
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def cfe_future_data_download(self, start_date):
        try:
            month_part = start_date[:6]
            date_part = start_date[6:]
            url = f'http://www.cffex.com.cn/sj/hqsj/rtj/{month_part}/{date_part}/index.xml?id=15'
            response = requests.get(url)

            tag = response._content.decode('utf8').replace('<?xml version="1.0" encoding="UTF-8"?>\n\n<dailydatas>','')
            data = tag.splitlines()
            instrumentid = []
            tradingday = []
            openprice = []
            highestprice = []
            lowestprice = []
            closeprice = []
            preopeninterest = []
            openinterest = []
            presettlementprice = []
            settlementprice = []
            volume = []
            turnover = []
            expiredate = []
            # 股指期货数据
            for line in data:
                if 'IO' in line:
                    break
                else:
                    if 'instrumentid' in line:
                        instrumentid.append(line.split(' ')[-1].partition("instrumentid>")[-1].partition("</instrumentid")[0])
                    elif 'tradingday' in line:
                        tradingday.append(line.split(' ')[-1].partition("tradingday>")[-1].partition("</tradingday")[0])
                    elif 'openprice' in line:
                        openprice.append(line.split(' ')[-1].partition("openprice>")[-1].partition("</openprice")[0])
                    elif 'highestprice' in line:
                        highestprice.append(line.split(' ')[-1].partition("highestprice>")[-1].partition("</highestprice")[0])
                    elif 'lowestprice' in line:
                        lowestprice.append(line.split(' ')[-1].partition("lowestprice>")[-1].partition("</lowestprice")[0])
                    elif 'closeprice' in line:
                        closeprice.append(line.split(' ')[-1].partition("closeprice>")[-1].partition("</closeprice")[0])
                    elif 'preopeninterest' in line:
                        preopeninterest.append(line.split(' ')[-1].partition("preopeninterest>")[-1].partition("</preopeninterest")[0])
                    elif 'openinterest' in line:
                        openinterest.append(line.split(' ')[-1].partition("openinterest>")[-1].partition("</openinterest")[0])
                    elif 'presettlementprice' in line:
                        presettlementprice.append(line.split(' ')[-1].partition("presettlementprice>")[-1].partition("</presettlementprice")[0])
                    elif 'settlementprice' in line:
                        settlementprice.append(line.split(' ')[-1].partition("settlementprice>")[-1].partition("</settlementprice")[0])
                    elif 'volume' in line:
                        volume.append(line.split(' ')[-1].partition("volume>")[-1].partition("</volume")[0])
                    elif 'turnover' in line:
                        turnover.append(line.split(' ')[-1].partition("turnover>")[-1].partition("</turnover")[0])
                    elif 'expiredate' in line:
                        expiredate.append(line.split(' ')[-1].partition("expiredate>")[-1].partition("</expiredate")[0])

                    else:
                        pass
            # 国债期货数据
            acount = 0
            for line in data:
                if 'T' in line:
                    acount = 1
                else:
                    pass
                if acount == 1:
                    if 'instrumentid' in line:
                        instrumentid.append(line.split(' ')[-1].partition("instrumentid>")[-1].partition("</instrumentid")[0])
                    elif 'tradingday' in line:
                        tradingday.append(line.split(' ')[-1].partition("tradingday>")[-1].partition("</tradingday")[0])
                    elif 'openprice' in line:
                        openprice.append(line.split(' ')[-1].partition("openprice>")[-1].partition("</openprice")[0])
                    elif 'highestprice' in line:
                        highestprice.append(line.split(' ')[-1].partition("highestprice>")[-1].partition("</highestprice")[0])
                    elif 'lowestprice' in line:
                        lowestprice.append(line.split(' ')[-1].partition("lowestprice>")[-1].partition("</lowestprice")[0])
                    elif 'closeprice' in line:
                        closeprice.append(line.split(' ')[-1].partition("closeprice>")[-1].partition("</closeprice")[0])
                    elif 'preopeninterest' in line:
                        preopeninterest.append(line.split(' ')[-1].partition("preopeninterest>")[-1].partition("</preopeninterest")[0])
                    elif 'openinterest' in line:
                        openinterest.append(line.split(' ')[-1].partition("openinterest>")[-1].partition("</openinterest")[0])
                    elif 'presettlementprice' in line:
                        presettlementprice.append(line.split(' ')[-1].partition("presettlementprice>")[-1].partition("</presettlementprice")[0])
                    elif 'settlementprice' in line:
                        settlementprice.append(line.split(' ')[-1].partition("settlementprice>")[-1].partition("</settlementprice")[0])
                    elif 'volume' in line:
                        volume.append(line.split(' ')[-1].partition("volume>")[-1].partition("</volume")[0])
                    elif 'turnover' in line:
                        turnover.append(line.split(' ')[-1].partition("turnover>")[-1].partition("</turnover")[0])
                    elif 'expiredate' in line:
                        expiredate.append(line.split(' ')[-1].partition("expiredate>")[-1].partition("</expiredate")[0])

                    else:
                        pass
                else:
                    pass

            # 结算价去空格（是个坑）
            clear = []
            for i in range(1,len(settlementprice),2):
                clear.append(settlementprice[i])
            dic = {'em_id':instrumentid ,'datetime':tradingday ,'open':openprice,'high':highestprice ,'low':lowestprice,
                'close':closeprice, 'uni_oi':openinterest,'pre_clear':presettlementprice ,
                'clear':clear,'uni_volume':volume,'uni_amount':turnover}

            tag = pd.DataFrame(dic)
            tag.datetime = pd.to_datetime(tag.datetime)
            tag.em_id = tag.em_id + '.CFE'
            self._data_helper._upload_raw(tag, EmFuturePriceDetail.__table__.name)
            return True

        except Exception as e:
            print(e)
            traceback.print_exc()
            return False


    def cfe_future_info_download(self):
        # 中金所期货信息下载
        try:
            # 国债期货部分信息
            def debt_info():
                dic1 ={'每日价格最大波动限制':'price_limit',
                    '最低交易保证金':'first_trans_margin',
                    '最后交易日':'last_trans_date',
                    '最小变动价位':'min_price_change',
                    '合约月份':'contract_date_info',
                    '交易时间':'trans_date_info',
                    '上市交易所':'mkt_name',
                    '交易代码':'product'}

                index_info = [['2ts','2017-02-17','2年期国债期货',20000,'元/手'],
                            ['5tf','2013-09-06','5年期国债期货',10000,'元/手'],
                            ['10tf','2015-03-20','10年期国债期货',10000,'元/手']]

                df1_ = pd.DataFrame()
                for index in index_info:
                    url_ = "http://www.cffex.com.cn/"+index[0]+"/"
                    f_ = requests.get(url_)                 
                    soup_ = BeautifulSoup(f_.content, "lxml")  


                    for k in soup_.find_all('div',class_='table_introduction debt_table_introduction'):
                        a = k.find_all('td')  

                    info_list = []

                    for line in a:
                        info_list.append(str(line).partition("td>")[-1].partition("</td")[0])

                    while '' in info_list:
                        info_list.remove('')
                    while '\n' in info_list:   
                        info_list.remove('\n')
                    key = []
                    value = []

                    for i in range(len(info_list)):
                        if i % 2==0:
                            key.append(info_list[i])
                        if i % 2==1:
                            value.append(info_list[i])

                    df=pd.DataFrame({'key':key,'value':value}).set_index('key').T
                    ddf2 = df[dic1.keys()].rename(columns=dic1)
                    ddf2['list_standard_date'] = index[1]
                    ddf2['trans_type'] = index[2]
                    ddf2['trans_unit'] = str(index[3]*100) + index[4]
                    ddf2['contract_mul'] = index[3]
                    ddf2['price_unit'] = index[4]
                    ddf2['price_limit_range'] = ddf2['price_limit']
                    df1_ = pd.concat([df1_,ddf2],axis=0)
                    df1_['short_eng_name'] = None
                    df1_['eng_name'] = None
                    df1_
                return df1_
            # 股指期货部分信息
            def index_info():
                dic = {'合约标的 ':'trans_type',
                    '每日价格最大波动限制':'price_limit',
                    '最低交易保证金':'first_trans_margin',
                    '最后交易日':'last_trans_date',
                    '合约乘数':'trans_unit',
                    '报价单位':'price_unit',
                    '最小变动价位':'min_price_change',
                    '交割日期':'deliver_date_info',
                    '合约月份':'contract_date_info',
                    '交易时间':'trans_date_info',
                    '上市交易所':'mkt_name',
                    '交易代码':'product'}

                df_ = pd.DataFrame()
                indexs = [['hs300','2010-04-16'],['zz500','2015-04-16'],['sz50','2015-04-16']]
                for index in indexs:
                    url = "http://www.cffex.com.cn/"+index[0]+"/"
                    f = requests.get(url)                 #Get该网页从而获取该html内容
                    soup = BeautifulSoup(f.content, "lxml")  #用lxml解析器解析该网页的内容, 好像f.text也是返回的html


                    for k in soup.find_all('div',class_='table_introduction debt_table_introduction'):
                        a = k.find_all('td')  
                    info_list = []

                    for line in a:
                        info_list.append(str(line).partition("td>")[-1].partition("</td")[0])

                    while '' in info_list:
                        info_list.remove('')
                    while '\n' in info_list:   
                        info_list.remove('\n')
                    key = []
                    value = []

                    for i in range(len(info_list)):
                        if i % 2==0:
                            key.append(info_list[i])
                        if i % 2==1:
                            value.append(info_list[i])

                    ddf=pd.DataFrame({'key':key,'value':value}).set_index('key').T
                    ddf['合约标的 '] = ddf['合约标的 '][0][:-2] + '股指期货'
                    ddf1 = ddf[dic.keys()].rename(columns=dic)
                    pattern = re.compile(r'\d+')
                    res = re.findall(pattern, ddf1['trans_unit'][0])
                    ddf1['contract_mul'] = res
                    ddf1['list_standard_date'] = index[1]
                    ddf1['price_limit_range'] = ddf1['price_limit']
                    df_ = pd.concat([df_,ddf1],axis=0)
                    df_['short_eng_name'] = None
                    df_['eng_name'] = None
                return df_

            # 基准价\开始\结束交易日
            url = 'http://www.cffex.com.cn/cp/index_6719.xml?id=1'
            response = requests.get(url)
            tag = response._content.decode('utf8').replace('<?xml version="1.0" encoding="UTF-8"?>\n\n<dailydatas>','')
            data = tag.splitlines()

            trading_code = []
            quote_price = []
            start_trade_date = []
            last_trade_date = []

            for line in data:
                if 'IO' in line:
                    break
                else:
                    if 'INSTRUMENTID' in line:
                        trading_code.append(line.split(' ')[-1].partition("INSTRUMENTID>")[-1].partition("</INSTRUMENTID")[0])
                    elif 'BASISPRICE' in line:
                        quote_price.append(line.split(' ')[-1].partition("BASISPRICE>")[-1].partition("</BASISPRICE")[0])
                    elif 'OPENDATE' in line:
                        start_trade_date.append(line.split(' ')[-1].partition("OPENDATE>")[-1].partition("</OPENDATE")[0])
                    elif 'ENDDELIVDATE' in line:
                        last_trade_date.append(line.split(' ')[-1].partition("ENDDELIVDATE>")[-1].partition("</ENDDELIVDATE")[0])
                    else:
                        pass

            acount = 0
            for line in data:
                if 'T2' in line:
                    acount = 1
                else:
                    pass
                if acount == 1:
                    if 'INSTRUMENTID' in line:
                        trading_code.append(line.split(' ')[-1].partition("INSTRUMENTID>")[-1].partition("</INSTRUMENTID")[0])
                    elif 'BASISPRICE' in line:
                        quote_price.append(line.split(' ')[-1].partition("BASISPRICE>")[-1].partition("</BASISPRICE")[0])
                    elif 'OPENDATE' in line:
                        start_trade_date.append(line.split(' ')[-1].partition("OPENDATE>")[-1].partition("</OPENDATE")[0])
                    elif 'ENDDELIVDATE' in line:
                        last_trade_date.append(line.split(' ')[-1].partition("ENDDELIVDATE>")[-1].partition("</ENDDELIVDATE")[0])
                    else:
                        pass
                else:
                    pass  
            df = pd.DataFrame({'trading_code':trading_code,
                            'quote_price':quote_price,
                            'start_trade_date':start_trade_date,
                            'last_trade_date':last_trade_date})

            df['deliv_month'] = df['last_trade_date'].apply(lambda x:x[:6])
            df.start_trade_date = pd.to_datetime(df.start_trade_date)
            df.last_trade_date = pd.to_datetime(df.last_trade_date)
            df['deliver_date'] = df['last_trade_date']

            # 保证金
            url1 = 'http://www.cffex.com.cn/sj/jscs/202108/16/index.xml?id=26'
            response1 = requests.get(url1)
            tag1 = response1._content.decode('utf8').replace('<?xml version="1.0" encoding="UTF-8"?>\n\n<dailydatas>','')
            data1 = tag1.splitlines()

            trading_code = []
            trans_margin = []
            long_margin = []
            short_margin = []
            product = []

            for line in data1:
                if 'INSTRUMENTID' in line:
                    trading_code.append(line.split(' ')[-1].partition("INSTRUMENTID>")[-1].partition("</INSTRUMENTID")[0])
                elif 'JGHMARGIN' in line:
                    trans_margin.append(line.split(' ')[-1].partition("JGHMARGIN>")[-1].partition("</JGHMARGIN")[0])
                elif 'LONGMARGINRATIO' in line:
                    long_margin.append(line.split(' ')[-1].partition("LONGMARGINRATIO>")[-1].partition("</LONGMARGINRATIO")[0])
                elif 'SHORTMARGINRATIO' in line:
                    short_margin.append(line.split(' ')[-1].partition("SHORTMARGINRATIO>")[-1].partition("</SHORTMARGINRATIO")[0])
                elif 'PRODUCTID' in line:
                    product.append(line.split(' ')[-1].partition("PRODUCTID>")[-1].partition("</PRODUCTID")[0])
                else:
                    pass

            df1 = pd.DataFrame({'trading_code':trading_code,
                                'trans_margin':trans_margin,
                                'long_margin':long_margin,
                                'short_margin':short_margin,
                                'product':product})
            df1[['long_margin','short_margin']] = df1[['long_margin','short_margin']].applymap(lambda x:float(x)*100)

            df1['em_id'] = df1.trading_code+'.CFE'
            df1['desc_name'] = df1.trading_code
            # 最后合并表格
            df_=index_info()
            df1_ = debt_info()

            result1 = pd.merge(pd.merge(df1,df_,on='product',how='inner'),df,on='trading_code')
            result2 = pd.merge(pd.merge(df1,df1_,on='product',how='inner'),df,on='trading_code')
            result = pd.concat([result1,result2],axis=0)
            result = result.drop(columns='product')
            exsited_future_info = RawDataApi().get_em_future_info()
            new_df = result[~result.em_id.isin(exsited_future_info.em_id)]
            if not new_df.empty:
                self._data_helper._upload_raw(new_df, EmFutureInfoDetail.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def dce_future_info_download(self):
        try:
            future = [['铁矿石','http://www.dce.com.cn/dalianshangpin/sspz/487477/487481/1500303/index.html','2013-05-16',2,32],
                    ['焦煤','http://www.dce.com.cn/dalianshangpin/sspz/487450/487454/1500228/index.html','2013-03-22',0,30],
                    ['焦炭','http://www.dce.com.cn/dalianshangpin/sspz/487423/487427/1499525/index.html','2019-04-30',2,32],
                    ['聚乙烯','http://www.dce.com.cn/dalianshangpin/sspz/487342/487346/1498988/index.html','2007-07-31',1,31],
                    ['聚氯乙烯','http://www.dce.com.cn/dalianshangpin/sspz/487369/487373/1499103/index.html','2009-05-25',1,31],
                    ['聚丙烯','http://www.dce.com.cn/dalianshangpin/sspz/487396/487400/1499431/index.html','2014-02-28',1,31],
                    ['乙二醇','http://www.dce.com.cn/dalianshangpin/sspz/yec63/6138350/6139005/index.html','2018-12-10',2,32],
                    ['苯乙烯','http://www.dce.com.cn/dalianshangpin/sspz/byx/hyygz42/6189178/index.html','2019-09-26',0,30],
                    ['液化石油气','http://www.dce.com.cn/dalianshangpin/sspz/yhsyq/hyygz7622/6210766/index.html','2020-03-30',1,31],
                    ['玉米','http://www.dce.com.cn/dalianshangpin/sspz/ym/hyygz/486238/index.html','2004-09-22',0,30],
                    ['淀粉','http://www.dce.com.cn/dalianshangpin/sspz/2081746/2081749/2081867/index.html','2014-12-19',1,31],
                    ['豆一','http://www.dce.com.cn/dalianshangpin/sspz/487124/487128/1391009/index.html','1999-01-18',1,31],
                    ['豆二','http://www.dce.com.cn/dalianshangpin/sspz/487153/487157/6040854/index.html','2004-12-22',2,32],
                    ['豆粕','http://www.dce.com.cn/dalianshangpin/sspz/487180/487184/1391326/index.html','2000-07-17',2,32],
                    ['豆油','http://www.dce.com.cn/dalianshangpin/sspz/487207/487211/1472342/index.html','2006-01-09',2,32],
                    ['棕榈油','http://www.dce.com.cn/dalianshangpin/sspz/487234/487238/1495874/index.html','2007-10-29',2,32],
                    ['鸡蛋','http://www.dce.com.cn/dalianshangpin/sspz/487261/487265/1498768/index.html','2013-11-08',2,32],
                    ['粳米','http://www.dce.com.cn/dalianshangpin/sspz/jm43/hyygz18/6178989/index.html','2019-08-16',2,32],
                    ]
            tag = pd.DataFrame()
            for index in future:
            #     print(index[0])
                url = index[1]
                f = requests.get(url)                 
                soup = BeautifulSoup(f.content, "lxml")  
                for k in soup.find_all('div',class_='detail_content'):
                    a = k.find_all('p')  
                info_list = []
                for line in a:
                    info_list.append(str(line).partition(';">')[-1].partition("</p")[0])
                    
                for i in range(len(info_list)):
                    if 'DCE' in info_list[i]:
                        info_list[i] = info_list[i].partition('（F')[0]

                while '' in info_list:
                    info_list.remove('') 
                        
                info_list = info_list[index[3]:index[4]] 
                key = []
                value = []
                for i in range(len(info_list)):
                    if i % 2==0:
                        key.append(info_list[i])
                    if i % 2==1:
                        value.append(info_list[i])
                ddf=pd.DataFrame({'key':key,'value':value}).set_index('key').T
                dic = {'涨跌停板幅度':'price_limit',
                    '最低交易保证金':'first_trans_margin',
                    '最后交易日':'last_trans_date',
                    '交易单位':'trans_unit',
                    '报价单位':'price_unit',
                    '最小变动价位':'min_price_change',
                    '最后交割日':'deliver_date_info',
                    '合约月份':'contract_date_info',
                    '交易时间':'trans_date_info',
                    '上市交易所':'mkt_name',
                    '交易代码':'product'}
                ddf1 = ddf[dic.keys()].rename(columns=dic)
                pattern = re.compile(r'\d+')
                res = re.findall(pattern, ddf1['trans_unit'][0])
                ddf1['contract_mul'] = res
                res1 = re.findall(pattern, ddf1['price_limit'][0])
                ddf1['price_limit_range'] = res1
                res2 = re.findall(pattern, ddf1['first_trans_margin'][0])
                ddf1['trans_margin'] = res2
                ddf1['long_margin']= res2
                ddf1['short_margin']= res2
                ddf1['list_standard_date'] = index[2]
                tag = pd.concat([tag,ddf1],axis=0)
            # 获取表格
            url1 = 'http://www.dce.com.cn/publicweb/businessguidelines/queryContractInfo.html'
            f1 = requests.get(url1)  
            soup1 = BeautifulSoup(f1.content, "lxml")
            for k in soup1.find_all('div',class_='dataWrapper'):
                c = k.find_all('tr')

            info_list1 = []
            for line in c:
                info_list1.append(str(line).partition('>')[-1].partition("</tr")[0])    
                
            lis1 = []
            for n in info_list1[0].split('><'):
                lis1.append(str(n).partition('>')[-1].partition("</th")[0])
                
            info = pd.DataFrame()
            for i in range(1,len(info_list1)):
                lis = []
                for n in info_list1[i].split('><'):
                    lis.append(str(n).partition('>')[-1].partition("</td")[0])

                pattern = re.compile(r'\d+')
                lis[3] = re.findall(pattern, info_list1[i].split('><')[3])[0]
                info = pd.concat([info,pd.DataFrame(lis)],axis=1) 
            df = pd.concat([pd.DataFrame({'key':lis1}),info],axis=1).set_index('key').T.reset_index(drop=True)
                
            dic1 = {'品种':'trans_type',
                    '合约代码':'trading_code',
                    '开始交易日':'start_trade_date',
                    '最后交易日':'last_trade_date',
                    '最后交割日':'deliver_date'}
            df = df[dic1.keys()].rename(columns=dic1)
            df['product'] = df['trading_code'].apply(lambda x:x[:-4].upper())
            df['em_id'] = df['trading_code'].apply(lambda x:x.upper() + '.DCE')
            result = pd.merge(df,tag,on='product',how='inner').drop(columns='product')
            result['deliv_month'] = result['deliver_date'].apply(lambda x:x[:6])
            result['start_trade_date'] = result['start_trade_date'].apply(lambda x:pd.to_datetime(str(x),format='%Y-%m-%d'))
            result['last_trade_date'] = result['last_trade_date'].apply(lambda x:pd.to_datetime(str(x),format='%Y-%m-%d'))
            result['deliver_date'] = result['deliver_date'].apply(lambda x:pd.to_datetime(str(x),format='%Y-%m-%d'))
            result['desc_name'] = result['trans_type'] + result['trading_code'].apply(lambda x:x[-4:])
            result['short_eng_name'] = None
            result['eng_name'] = None
            result['quote_price'] = None
            result['price_limit'] = result['price_limit'].map(lambda x: ''.join([s for s in x if s.isdigit()]))
            exsited_future_info = RawDataApi().get_em_future_info()
            new_df = result[~result.em_id.isin(exsited_future_info.em_id)]
            if not new_df.empty:
                self._data_helper._upload_raw(new_df, EmFutureInfoDetail.__table__.name)
            return True

        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def download_all(self, start_date, end_date):
        failed_tasks = []
        if not self.cm_index_price(start_date, end_date):
            failed_tasks.append('cm_index_price')

        if not self.cxindex_index_price(start_date, end_date):
            failed_tasks.append('cxindex_index_price')

        if not self.yahoo_index_price(start_date, end_date):
            failed_tasks.append('yahoo_index_price')

        if not self.sina_btc_download():
            failed_tasks.append('sina_btc_close')

        return failed_tasks


if __name__ == "__main__":
    data_helper = RawDataHelper()
    web_downloader = WebRawDataDownloader(data_helper)
    # web_downloader.cm_index_price('20200408', '20200508')
    # web_downloader.yahoo_index_price('20210709', '20210709')
    # web_downloader.cxindex_index_price('20210617', '20210617')
